@@ -19,9 +19,9 @@
 
 """Provides the FocusTrackingPresenter for Orca."""
 
-__id__  = "$Id: focus_tracking_presenter.py 4088 2008-08-06 20:42:48Z wwalker $"
-__version__   = "$Revision: 4088 $"
-__date__      = "$Date: 2008-08-06 16:42:48 -0400 (Wed, 06 Aug 2008) $"
+__id__  = "$Id: focus_tracking_presenter.py 4445 2009-01-21 15:18:00Z wwalker $"
+__version__   = "$Revision: 4445 $"
+__date__      = "$Date: 2009-01-21 10:18:00 -0500 (Wed, 21 Jan 2009) $"
 __copyright__ = "Copyright (c) 2005-2008 Sun Microsystems Inc."
 __license__   = "LGPL"
 
@@ -35,6 +35,7 @@ import braille
 import default
 import debug
 import input_event
+import orca
 import orca_state
 import presentation_manager
 import settings
@@ -543,6 +544,13 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                 if state.contains(pyatspi.STATE_DEFUNCT):
                     debug.println(debug.LEVEL_FINEST,
                                   "IGNORING DEFUNCT OBJECT")
+                    if event.type.startswith("window:deactivate"):
+                        if orca_state.activeScript \
+                           and orca_state.activeScript.flatReviewContext:
+                            orca_state.activeScript.drawOutline(-1, 0, 0, 0)
+                            orca_state.activeScript.flatReviewContext = None
+                        orca.setLocusOfFocus(event, None)
+                        orca_state.activeWindow = None
                     return
 
             if (not debug.eventDebugFilter) \
