@@ -23,16 +23,16 @@ implementation in Gecko:
 http://developer.mozilla.org/en/docs/Accessibility/ATSPI_Support
 """
 
-__id__        = "$Id$"
-__version__   = "$Revision$"
-__date__      = "$Date$"
+__id__        = "$Id: speech_generator.py 4535 2009-02-13 00:55:11Z wwalker $"
+__version__   = "$Revision: 4535 $"
+__date__      = "$Date: 2009-02-12 19:55:11 -0500 (Thu, 12 Feb 2009) $"
 __copyright__ = "Copyright (c) 2005-2008 Sun Microsystems Inc."
 __license__   = "LGPL"
 
-import math
 import pyatspi
 
 import orca.rolenames as rolenames
+import orca.settings as settings
 import orca.speechgenerator as speechgenerator
 
 from orca.orca_i18n import _
@@ -64,16 +64,52 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
 
     def getSpeechForObjectRole(self, obj, role=None):
         """Prevents some roles from being spoken."""
-        if obj.getRole() in [pyatspi.ROLE_PARAGRAPH,
-                             pyatspi.ROLE_SECTION,
-                             pyatspi.ROLE_LABEL,
-                             pyatspi.ROLE_FORM,
-                             pyatspi.ROLE_LIST_ITEM,
-                             pyatspi.ROLE_MENU_ITEM,
-                             pyatspi.ROLE_UNKNOWN]:
-            return []
+        doNotSpeak = [pyatspi.ROLE_FORM,
+                      pyatspi.ROLE_LABEL,
+                      pyatspi.ROLE_MENU_ITEM,
+                      pyatspi.ROLE_PARAGRAPH,
+                      pyatspi.ROLE_SECTION,
+                      pyatspi.ROLE_UNKNOWN]
+
+        if self._script.inDocumentContent(obj):
+            doNotSpeak.append(pyatspi.ROLE_TABLE_CELL)
+            if not self._script.isAriaWidget(obj):
+                doNotSpeak.append(pyatspi.ROLE_LIST_ITEM)
+                doNotSpeak.append(pyatspi.ROLE_LIST)
+
+        utterances = []
+        if role or not obj.getRole() in doNotSpeak:
+            utterances.append(rolenames.getSpeechForRoleName(obj, role))
+            if obj.getRole() == pyatspi.ROLE_HEADING:
+                level = self._script.getHeadingLevel(obj)
+                if level:
+                    # Translators: this is in reference to a heading level
+                    # in HTML (e.g., For <h3>, the level is 3).
+                    #
+                    utterances.append(_("level %d") % level)
+
+        return utterances
+
+    def _getSpeechForAlert(self, obj, already_focused):
+        """Gets the speech for an alert.  What we do here is first try
+        to see if the alert contains text via embedded object characters.
+        If it does, we speak that text.  If it doesn't, we defer to the
+        super class.  The prototype alert we're shooting for is the one
+        from http://bugzilla.gnome.org/show_bug.cgi?id=570551
+
+        Arguments:
+        - obj: an Accessible
+        - already_focused: False if object just received focus
+
+        Returns a list of utterances to be spoken for the object.
+        """
+
+        text = self._script.expandEOCs(obj)
+        if text:
+            return [text]
         else:
-            return [rolenames.getSpeechForRoleName(obj, role)]
+            return speechgenerator.SpeechGenerator.\
+                       _getSpeechForAlert(self, obj, already_focused)
 
     def _getSpeechForDocumentFrame(self, obj, already_focused):
         """Gets the speech for a document frame.
@@ -142,6 +178,10 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
             return speechgenerator.SpeechGenerator._getSpeechForText(
                 self, obj, already_focused)
 
+        if settings.presentReadOnlyText \
+           and self._script.isReadOnlyTextArea(obj):
+            utterances.append(settings.speechReadOnlyString)
+
         utterances.extend(self.getSpeechForObjectRole(obj))
 
         [text, caretOffset, startOffset] = self._script.getTextLineAtCaret(obj)
@@ -198,6 +238,11 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
                 # This should work...
                 #
                 child = menu.querySelection().getSelectedChild(0)
+                if not child:
+                    # It's probably a Gtk combo box.
+                    #
+                    return speechgenerator.SpeechGenerator.\
+                        _getSpeechForComboBox(self, obj, already_focused)
             except:
                 # But just in case, we'll fall back on this.
                 # [[[TODO - JD: Will we ever have a case where the first
@@ -317,28 +362,27 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
 
         utterances = []
 
-        if obj.getState().contains(pyatspi.STATE_FOCUSED):
-            label = self._script.getDisplayedLabel(obj)
-            if not label:
-                if not self._script.inDocumentContent():
-                    label = obj.name
-                else:
-                    label = self._script.guessTheLabel(obj)
+        label = self._script.getDisplayedLabel(obj)
+        if not label:
+            if not self._script.inDocumentContent():
+                label = obj.name
+            else:
+                label = self._script.guessTheLabel(obj)
 
-            if not already_focused and label:
-                utterances.append(label)
+        if not already_focused and label:
+            utterances.append(label)
 
-            item = None
-            selection = obj.querySelection()
-            for i in xrange(obj.childCount):
-                if selection.isChildSelected(i):
-                    item = obj[i]
-                    break
-            item = item or obj[0]
-            if item:
-                name = self._getSpeechForObjectName(item)
-                if name != label:
-                    utterances.extend(name)
+        item = None
+        selection = obj.querySelection()
+        for i in xrange(obj.childCount):
+            if selection.isChildSelected(i):
+                item = obj[i]
+                break
+        item = item or obj[0]
+        if item:
+            name = self._getSpeechForObjectName(item)
+            if name != label:
+                utterances.extend(name)
 
         if not already_focused:
             if obj.getState().contains(pyatspi.STATE_MULTISELECTABLE):
@@ -396,12 +440,11 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
             link = self._script.getAncestor(obj,
                                             [pyatspi.ROLE_LINK],
                                             [pyatspi.ROLE_DOCUMENT_FRAME])
-            if not len(utterances):
-                if link:
-                    utterances.extend(self._getSpeechForLink(link,
-                                                             already_focused))
-            elif link:
-                utterances.extend(self.getSpeechForObjectRole(link))
+            if link:
+                if not len(utterances):
+                    return self._getSpeechForLink(link, already_focused)
+                else:
+                    utterances.extend(self.getSpeechForObjectRole(link))
 
             utterances.extend(self.getSpeechForObjectRole(obj))
 
@@ -424,12 +467,6 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
         Returns a list of utterances to be spoken for the object.
         """
 
-        # Treat ARIA widgets like default.py widgets
-        #
-        if self._script.isAriaWidget(obj):
-            return speechgenerator.SpeechGenerator.getSpeech(obj,
-                                                             already_focused)
-        
         utterances = []
 
         if not already_focused:
@@ -454,6 +491,12 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
 
             utterances.extend(self.getSpeechForObjectRole(obj))
 
+            # If the link has a child which is an image, we want
+            # to indicate that.
+            #
+            if obj.childCount and obj[0].getRole() == pyatspi.ROLE_IMAGE:
+                utterances.extend(self.getSpeechForObjectRole(obj[0]))
+
         utterances.extend(self._getSpeechForObjectAvailability(obj))
 
         self._debugGenerator("Gecko._getSpeechForLink",
@@ -630,12 +673,17 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
 
         # Skip items of unknown rolenames, menu bars, labels with 
         # children, and autocompletes.  (With autocompletes, we
-        # wind up speaking the text object)
+        # wind up speaking the text object). Beginning with Firefox
+        # 3.2, list items have names corresponding with their text.
+        # This results in getDisplayedText returning actual text
+        # and the parent list item being spoken when it should not
+        # be. So we'll take list items out of the context.
         #
         skipRoles = [pyatspi.ROLE_UNKNOWN,
                      pyatspi.ROLE_MENU_BAR,
                      pyatspi.ROLE_LABEL,
-                     pyatspi.ROLE_AUTOCOMPLETE]
+                     pyatspi.ROLE_AUTOCOMPLETE,
+                     pyatspi.ROLE_LIST_ITEM]
 
         # Stop if we get to a document frame or an internal frame.
         #
@@ -670,12 +718,21 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
             # Also skip the parent if its accessible text is a single 
             # EMBEDDED_OBJECT_CHARACTER: Script.getDisplayedText will
             # end up coming back to the child of an object for the text
-            # if an object's text contains a single EOC.
+            # if an object's text contains a single EOC. In addition,
+            # beginning with Firefox 3.2, a table cell may derive its
+            # accessible name from focusable objects it contains (e.g.
+            # links, form fields). getDisplayedText will return the
+            # object's name in this case (because of the presence of
+            # the EOC and other characters). This causes us to be
+            # chatty. So if it's a table cell which contains an EOC,
+            # we will also skip the parent.
             #
             parentText = self._script.queryNonEmptyText(parent)
             if parentText:
                 unicodeText = parentText.getText(0, -1).decode("UTF-8")
-                if unicodeText == self._script.EMBEDDED_OBJECT_CHARACTER:
+                if self._script.EMBEDDED_OBJECT_CHARACTER in unicodeText \
+                   and (len(unicodeText) == 1 \
+                        or role == pyatspi.ROLE_TABLE_CELL):
                     parent = parent.parent
                     continue
 
@@ -696,6 +753,13 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
             if not role in dontSpeakRoles and len(newUtterances):
                 utterances.append(rolenames.getSpeechForRoleName(parent))
 
+            # If this object is an ARIA widget with STATE_REQUIRED, add
+            # that. (Note that for the most part, the ARIA widget itself
+            # has this state, but in the case of a group of radio buttons,
+            # it is the group which has the state).
+            #
+            utterances.extend(self._getSpeechForRequiredObject(parent))
+
             utterances.extend(newUtterances)
 
             parent = parent.parent
@@ -716,44 +780,11 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
         """
         
         # Let default handle non-ARIA widgets (XUL?)
-        if not self._script.isAriaWidget(obj):
+        if self._script.isAriaWidget(obj):
             return speechgenerator.SpeechGenerator.\
                        _getSpeechForSlider(self, obj, already_focused)
 
-        value = obj.queryValue()
-
-        # OK, this craziness is all about trying to figure out the most
-        # meaningful formatting string for the floating point values.
-        # The number of places to the right of the decimal point should
-        # be set by the minimumIncrement, but the minimumIncrement isn't
-        # always set.  So...we'll default the minimumIncrement to 1/100
-        # of the range.  But, if max == min, then we'll just go for showing
-        # them off to two meaningful digits.
-        #
-        try:
-            minimumIncrement = value.minimumIncrement
-        except:
-            minimumIncrement = 0.0
-
-        if minimumIncrement == 0.0:
-            minimumIncrement = (value.maximumValue - value.minimumValue) \
-                               / 100.0
-
-        try:
-            decimalPlaces = max(0, -math.log10(minimumIncrement))
-        except:
-            try:
-                decimalPlaces = max(0, -math.log10(value.minimumValue))
-            except:
-                try:
-                    decimalPlaces = max(0, -math.log10(value.maximumValue))
-                except:
-                    decimalPlaces = 0
-
-        formatter = "%%.%df" % decimalPlaces
-        valueString = formatter % value.currentValue
-        #minString   = formatter % value.minimumValue
-        #maxString   = formatter % value.maximumValue
+        valueString = self._script.getTextForValue(obj)
 
         if already_focused:
             utterances = [valueString]
@@ -764,7 +795,7 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
             utterances.append(valueString)
             utterances.extend(self._getSpeechForObjectAvailability(obj))
 
-        self._debugGenerator("_getSpeechForSlider",
+        self._debugGenerator("Gecko._getSpeechForSlider",
                              obj,
                              already_focused,
                              utterances)
